SUBROUTINE check_fetch_order(nbpt, nbbas, fetch, outgrid, outbas, basin_count)

  IMPLICIT NONE

  !! INPUT VARIABLES
  INTEGER, INTENT(in) :: nbpt, nbbas
  REAL,DIMENSION(nbpt, nbbas), INTENT(in) :: fetch
  INTEGER, DIMENSION(nbpt, nbbas), INTENT(in) :: outgrid, outbas
  INTEGER, DIMENSION(nbpt), INTENT(in) :: basin_count

  !! LOCAL VARIABLES
  INTEGER :: ig, ib, ibmx
  REAL :: f1, f2
  

  WRITE(*,*) " *** Checking fetch order errors *** "
  DO ig=1,nbpt
    ibmx = basin_count(ig)
    DO ib=1,ibmx
      IF (outbas(ig,ib) .LE. 35) THEN
        f1 = fetch(ig,ib)
        f2 = fetch(outgrid(ig,ib), outbas(ig,ib))
        IF (f1 .GT. f2) THEN 
          WRITE(*,*) "*ERROR*", ig, ib, f1, f2
          WRITE(*,*) "outflows : ", outgrid(ig,ib), outbas(ig,ib)
        END IF
      END IF
    END DO
  END DO 

END SUBROUTINE check_fetch_order

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE check_inflows_in_outflow(nbpt, nbbas, nbin, outgrid, outbas, basin_count, ingrid, inbas, innum)

  IMPLICIT NONE

  !! INPUT VARIABLES
  INTEGER, INTENT(in) :: nbpt, nbbas, nbin
  INTEGER, DIMENSION(nbpt, nbbas), INTENT(in) :: outgrid, outbas, innum
  INTEGER, DIMENSION(nbpt, nbbas, nbin), INTENT(in) :: ingrid, inbas
  INTEGER, DIMENSION(nbpt), INTENT(in) :: basin_count

  !! LOCAL VARIABLES
  INTEGER :: ig, ib, ibmx, og, ob, test
  INTEGER :: iin, iinmx, ii, limiin


  WRITE(*,*) " *** Checking upstream HTUs is in the inflows *** "

  DO ig=1,nbpt
    ibmx = basin_count(ig)
    DO ib=1,ibmx
      og = outgrid(ig,ib)
      ob = outbas(ig,ib)
      IF ( ob .LE. 35) THEN
        iinmx = innum(og,ob)
        test = 0
        ! Look through the inflows
        DO iin=1,iinmx
          IF (test .EQ. 0) THEN
              IF ((ingrid(og,ob,iin) .EQ. ig) .AND. (inbas(og,ob,iin) .EQ. ib)) test = 1
              IF (ingrid(og,ob,iin) .EQ. 0) THEN
                test = -1
                limiin = iin-1
              END IF
          END IF
        ENDDO
        ! Error : not present
        IF (test .EQ. 0) THEN 
          WRITE(*,*) "Error at", ig, ib
          WRITE(*,*) "Outflow:", og, ob, "with", iinmx, "inflows"
          DO ii=1,iinmx
            WRITE(*,*) ii, ingrid(og,ob,ii), inbas(og,ob,ii)
          END DO
        END IF
        ! Error, less inflows than innum (and not present)
        IF (test .EQ. -1) WRITE(*,*) "innum does not correspond (iin, iinmax)", limiin, iinmx
      END IF 

    END DO
  END DO 

END SUBROUTINE check_inflows_in_outflow

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!

SUBROUTINE check_outflow_in_inflows(nbpt, nbbas, nbin, outgrid, outbas, basin_count, ingrid, inbas, innum)

  IMPLICIT NONE

  !! INPUT VARIABLES
  INTEGER, INTENT(in) :: nbpt, nbbas, nbin
  INTEGER, DIMENSION(nbpt, nbbas), INTENT(in) :: outgrid, outbas, innum
  INTEGER, DIMENSION(nbpt, nbbas, nbin), INTENT(in) :: ingrid, inbas
  INTEGER, DIMENSION(nbpt), INTENT(in) :: basin_count

  !! LOCAL VARIABLES
  INTEGER :: ig, ib, ibmx, test, ing, inb
  INTEGER :: iinmx, ii


  WRITE(*,*) " *** Checking if the inflows corresponds to the upstream outflows  *** "

  DO ig=1,nbpt
    ibmx = basin_count(ig)
    DO ib=1,ibmx
      iinmx = innum(ig,ib)
      IF ( iinmx .GT. 0) THEN
        test = 0
        DO ii=1,iinmx
          ing = ingrid(ig,ib,ii)
          inb = inbas(ig,ib,ii)
          IF ((outgrid(ing,inb) .NE. ig) .OR. (outbas(ing,inb) .NE. ib)) THEN
            WRITE(*,*) "Error at", ig, ib
            WRITE(*,*) "outflow->", outgrid(ing,inb), outbas(ing,inb)
          END IF
        ENDDO
      END IF 
    END DO
  END DO 

END SUBROUTINE check_outflow_in_inflows

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!

SUBROUTINE test_outflow(nbpt, nbbas, outgrid, outbas, basin_count)

  IMPLICIT NONE
  ! Changes to do : 
  ! Optimize by keeping in memory the points already checked 

  !! INPUT VARIABLES
  INTEGER, INTENT(in) :: nbpt, nbbas
  INTEGER, DIMENSION(nbpt, nbbas), INTENT(in) :: outgrid, outbas
  INTEGER, DIMENSION(nbpt), INTENT(in) :: basin_count

  !! LOCAL VARIABLES
  INTEGER :: ig, ib, ibmx, og, ob, ind
  INTEGER :: it

  WRITE(*,*) " *** Checking all HTUs goes to ocean *** "

  DO ig=1,nbpt
    ibmx = basin_count(ig)
    DO ib=1,ibmx
      ind = 0
      og = outgrid(ig,ib)
      ob = outbas(ig,ib)
      DO WHILE (ob .LE. 35)
        it = outgrid(og,ob)
        ob = outbas(og,ob)
        og = it
        ind = ind+1
        IF (ind .GE. 5000) THEN
          og = 0
          WRITE(*,*) "ERROR", ig, ib
        END IF
      END DO
    END DO
  END DO


END SUBROUTINE test_outflow


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!


SUBROUTINE compute_fetch(nbpt, nbbas, outgrid, outbas, basin_count, basin_area, fetch_new)

  IMPLICIT NONE

  !! INPUT VARIABLES
  INTEGER, INTENT(in) :: nbpt, nbbas
  INTEGER, DIMENSION(nbpt, nbbas), INTENT(in) :: outgrid, outbas
  REAL, DIMENSION(nbpt, nbbas), INTENT(in) :: basin_area
  INTEGER, DIMENSION(nbpt), INTENT(in) :: basin_count

  !! OUTPUT VARIABLES
  REAL, DIMENSION(nbpt,nbbas), INTENT(out) :: fetch_new

  !! LOCAL VARIABLES
  INTEGER :: ig, ib, ibmx, og, ob, ind
  INTEGER :: it
  REAL :: area

  WRITE(*,*) " *** Computing the fetch *** "

  fetch_new(:,:) = 0

  DO ig=1,nbpt
    ibmx = basin_count(ig)
    DO ib=1,ibmx
      ind = 0
      area = basin_area(ig,ib)
      fetch_new(ig,ib) = fetch_new(ig,ib) + area

      og = outgrid(ig,ib)
      ob = outbas(ig,ib)
      DO WHILE (ob .LE. 35)
        fetch_new(og,ob) = fetch_new(og,ob) + area
        it = outgrid(og,ob)
        ob = outbas(og,ob)
        og = it
      END DO
    END DO
  END DO


END SUBROUTINE compute_fetch



SUBROUTINE compute_inflows(nbpt, nbbas, nbin, outgrid, outbas, basin_count, ingrid_new, inbas_new, innum_new)

  IMPLICIT NONE

  !! INPUT VARIABLES
  INTEGER, INTENT(in) :: nbpt, nbbas, nbin
  INTEGER, DIMENSION(nbpt, nbbas), INTENT(in) :: outgrid, outbas
  INTEGER, DIMENSION(nbpt), INTENT(in) :: basin_count

  INTEGER, DIMENSION(nbpt, nbbas), INTENT(OUT) :: innum_new
  INTEGER, DIMENSION(nbpt, nbbas, nbin), INTENT(OUT) :: ingrid_new, inbas_new

  !! LOCAL VARIABLES
  INTEGER :: ig, ib, ibmx, og, ob
  INTEGER :: iin

  innum_new(:,:)  = 0
  ingrid_new(:,:,:) = 0
  inbas_new(:,:,:)  = 0

  WRITE(*,*) " *** Computing the inflows  *** "

  DO ig=1,nbpt
    ibmx = basin_count(ig)
    DO ib=1,ibmx
      og = outgrid(ig,ib)
      ob = outbas(ig,ib)
      IF ( ob .LE. 35) THEN
        innum_new(og,ob) = innum_new(og,ob) +1
        iin = innum_new(og,ob)
        ingrid_new(og,ob,iin) = ig
        inbas_new(og,ob,iin) = ib
      END IF
    END DO
  END DO 

END SUBROUTINE compute_inflows




