import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap, cm
import numpy as np
from netCDF4 import Dataset
import sys

NCfile="../SPortugal_test_graph_n4.nc"
GraphicFile = "test.png"
title="From file : "+NCfile
#
#####################################################################
#
class LandGrid :
    def __init__(self, land) :
        nj,ni = land.shape
        self.indP=[]
        self.nbland = 0
        for i in range(ni) :
            for j in range(nj) :
                if (land[j,i] > 0 ) :
                    self.nbland = self.nbland+1
                    self.indP.append([j,i])
        return
                    
    def land2ij(self, il) :
        j,i = self.indP[il]
        return j,i
#
#####################################################################
#
#
nc = Dataset(NCfile, 'r')

lon_full = nc.variables["lon"][:,:]
lat_full = nc.variables["lat"][:,:]
lon_bnd = nc.variables["lon_bnd"][:,:,:]
lat_bnd = nc.variables["lat_bnd"][:,:,:]
land = nc.variables["land"][:,:]
routetogrid = nc.variables["routetogrid"][:,:,:]
routetohtu = nc.variables["routetobasin"][:,:,:]
#
# Coding for routetobasin :
# nbasmax + 1 = Returnflow to the grid
# nbasmax + 2 = Coastal flow
# nbasmax + 3 = River flow
#
cglon = nc.variables["CG_lon"][:,:,:]
cglat = nc.variables["CG_lat"][:,:,:]

nc.close()

nhtu, nj, ni = routetogrid.shape
landindex = LandGrid(land)
print("Nb land points = ", landindex.nbland, 'Nb HTU : ', nhtu)

#
# Get some average size of the grid boxes.
#
lon_res = np.nanmean(np.abs(np.gradient(lon_bnd, axis=0)))
lat_res = np.nanmean(np.abs(np.gradient(lat_bnd, axis=0)))
#
#
#
fig = plt.figure(1)
ax = fig.add_subplot(111)
ax.set_title(title)
ax.set_xlabel('Longitude',labelpad=20)
ax.set_ylabel('Latitude',labelpad=20)
ax.axis([0, 10, 0, 10])
#
m = Basemap(projection='cyl', resolution='i', \
            llcrnrlon=np.nanmin(lon_bnd)-lon_res, llcrnrlat=np.nanmin(lat_bnd)-lat_res, \
            urcrnrlon=np.nanmax(lon_bnd)+lon_res, urcrnrlat=np.nanmax(lat_bnd)+lat_res)

m.drawcoastlines(linewidth=0.5)
m.drawcountries(linewidth=0.5)
m.drawparallels(np.arange(-180.,180.,0.1),labels=[1,0,0,0]) # draw parallels
m.drawmeridians(np.arange(-80.,80.,0.1),labels=[0,0,0,1]) # draw meridians
#
cmap = 'RdBu'
cmap = 'Pastel1'
cmap = 'prism'
lwidth=0.0005
markersize=5
#
#
#
for i in range(ni) :
    for j in range(nj) :
        if land[j,i] > 0 :
            lo = np.append(lon_bnd[:,j,i],lon_bnd[0,j,i])
            la = np.append(lat_bnd[:,j,i],lat_bnd[0,j,i])
            xl,yl=m(lo, la)
            plt.plot(xl,yl, color="m")
#
# Draw the graph
#
for il in range(landindex.nbland) :
    j,i = landindex.land2ij(il)
    xt,yt=m(lon_full[j,i],lat_full[j,i])
    plt.text(xt,yt,str(il))
    act_nhtu=np.count_nonzero(routetohtu[:,j,i] > 0)
    for ih in range(act_nhtu) :
        lo = cglon[ih,j,i]
        la = cglat[ih,j,i]
        xl,yl=m(lo, la)
        if int(routetohtu[ih,j,i]-1) < nhtu :
            plt.plot(xl,yl, ms=markersize, marker=11)
        elif int(routetohtu[ih,j,i]-1) == nhtu+1 :
            plt.plot(xl,yl, ms=markersize, marker="x")
        elif int(routetohtu[ih,j,i]-1) == nhtu+2 :
            plt.plot(xl,yl, ms=markersize, marker="o")
        elif int(routetohtu[ih,j,i]-1) == nhtu+3 :
            plt.plot(xl,yl, ms=markersize, marker="p")
        else :
            print("Unforseen coding of outlof HTU : ",int(routetohtu[ih,j,i]-1))
            sys.exit()
        
        if int(routetohtu[ih,j,i]-1) < nhtu :
            ile =  int(routetogrid[ih,j,i]-1)
            je,ie = landindex.land2ij(ile)
            ihe = int(routetohtu[ih,j,i]-1)
            loe = cglon[ihe,je,ie]
            lae = cglat[ihe,je,ie]
            xle,yle=m(loe, lae)
            plt.arrow(xl, yl, xle-xl, yle-yl, color="b", width=lwidth, head_width=20*lwidth, length_includes_head=True)
#
#
#
plt.savefig(GraphicFile, dpi=450)
plt.close()
