
import numpy as np
from netCDF4 import Dataset

FillValue=1.0e20

def dumpnetcdf(filename, globalgrid, procgrid, part, x1dland, x1dxy, x2dland, fx) :
        #
        vtyp=np.float64
        cornerind=[0,2,4,6]
        #
        if part.rank == 0 :
                outnf=Dataset(filename, 'w', format='NETCDF4_CLASSIC')
                # Dimensions
                outnf.createDimension('lon', globalgrid.ni)
                outnf.createDimension('lat', globalgrid.nj)
                outnf.createDimension('land', globalgrid.nbland)
                outnf.createDimension('htu', part.size)
                outnf.createDimension('bnd', 4)
        #
        # Coordinates
        #
        # Longitude
        longitude = part.gather(procgrid.lon_full)
        if part.rank == 0 :
                lon=outnf.createVariable("lon", vtyp, ('lat','lon'), fill_value=FillValue)
                lon.units="grid box centre degrees east"
                lon.standard_name="grid longitude"
                lon.title="Longitude"
                lon.axis="X"
                lon[:,:] = longitude[:,:]
                # # Longitude bounds
                # lonbnd=outnf.createVariable("lon_bnd", vtyp, ('bnd','lat','lon'), fill_value=RPP.FillValue)
                # lonbnd.units="grid box corners degrees east"
                # lonbnd.title="Longitude Corners"
                # lonbnd[:,:] = np.array(modelgrid.polyll)[:,cornerind,0]
        #
        # Latitude
        latitude = part.gather(procgrid.lat_full)
        if part.rank == 0 :
                lat=outnf.createVariable("lat", vtyp, ('lat','lon'), fill_value=FillValue)
                lat.units="grid box centre degrees north"
                lat.standard_name="grid latitude"
                lat.title="Latitude"
                lat.axis="Y"
                lat[:,:] = latitude[:,:]
        #
        #
        # Land sea mask
        #
        if part.rank == 0 :
                land=outnf.createVariable("land", vtyp, ('lat','lon'), fill_value=FillValue)
                land.units="Land Sea mask"
                land.standard_name="landsea mask"
                land.title="Land"
                land[:,:] = globalgrid.land[:,:]
        #
        # Some variable
        #
        fullX = part.landscatgat(procgrid, x1dland)
        if part.rank == 0 :
                fullX[np.isnan(fullX)] = FillValue
                x1dl=outnf.createVariable("var1d", vtyp, ('lat','lon'), fill_value=FillValue)
                x1dl.units="x"
                x1dl.standard_name="Some land variable"
                x1dl.title="Land"
                x1dl[:,:] = fullX[:,:]
        #
        # Some 2D spatial variable
        #
        fullXY = part.gather(x1dxy)
        if part.rank == 0 :
                fullXY[np.isnan(fullXY)] = FillValue
                x1dxy=outnf.createVariable("var1dxy", vtyp, ('lat','lon'), fill_value=FillValue)
                x1dxy.units="x"
                x1dxy.standard_name="Some other land variable"
                x1dxy.title="Land"
                x1dxy[:,:] = fullXY[:,:]
        #
        # Some variable
        #
        fullX2D = part.landscatgat(procgrid, x2dland)
        if part.rank == 0 :
                fullX2D[np.isnan(fullX2D)] = FillValue
                x2dl=outnf.createVariable("var2d", vtyp, ('htu','lat','lon'), fill_value=FillValue)
                x2dl.units="x"
                x2dl.standard_name="Some 2D land variable"
                x2dl.title="2D Land"
                x2dl[:,:,:] = fullX2D[:,:,:]
        #
        # Test a Fortran gathered field
        #
        fg = procgrid.landscatter(fx[:,:], order='F')
        if part.rank == 0 :                           
            l2id = outnf.createVariable("land2d", vtyp, ('htu','lat','lon'), fill_value=FillValue)
            l2id.title = "Land 2d in F order"
            l2id.units = "-"
        else :
            l2id = np.zeros((1,1,1))
        l2id[:,:,:] = part.gather(fg)
        #
        # Proc by land
        if part.rank == 0 :
                outnf.close()

        return
