import numpy as np
import sys
#
import getargs
log_master, log_world = getargs.getLogger(__name__)
INFO, DEBUG, ERROR = log_master.info, log_master.debug, log_world.error
INFO_ALL, DEBUG_ALL = log_world.info, log_world.debug
#
def halfpartition(partin, land) :
    """
    First partition of the domain
    The domain with more land points is halved
    """
    A = [p["nbland"] for p in partin]
    i = np.argmax(A)
    partout = []
    partout = [partin[k] for k in range(len(partin)) if k != i]

    dom = partin[i]

    if dom["nbi"] > dom["nbj"] :
        hh = [h for h in range(dom["nbi"])]
        nb = np.array([np.ma.sum(np.ma.filled(land[dom["jstart"]:dom["jstart"]+dom["nbj"],dom["istart"]:dom["istart"]+h],0)) - dom['nbland']/2. for h in hh])
        i = np.nanargmin(np.abs(nb))
        new_dom = {"nbi":dom["nbi"]-hh[i],"nbj":dom["nbj"],  \
                   "istart":dom["istart"]+hh[i],"jstart":dom["jstart"]}
        dom["nbi"] = hh[i]
        
    else :
        hh = [h for h in range(dom["nbj"])]
        nb = np.array([np.ma.sum(np.ma.filled(land[dom["jstart"]:dom["jstart"]+h,dom["istart"]:dom["istart"]+dom["nbi"]],0)) - int(dom['nbland']/2.) for h in hh])
        i = np.nanargmin(np.abs(nb))
        new_dom = {"nbi":dom["nbi"],"nbj":dom["nbj"]-hh[i], \
                 "istart":dom["istart"],"jstart":dom["jstart"]+hh[i]}
        dom["nbj"] = hh[i]

    new_dom['nbland'] = int(np.nansum(land[new_dom["jstart"]:new_dom["jstart"]+new_dom["nbj"],new_dom["istart"]:new_dom["istart"]+new_dom["nbi"]]))
    dom['nbland'] = dom['nbland']-new_dom['nbland']

    partout.append(dom)
    partout.append(new_dom)

    return partout
#
# Divide in two the domain with the most number of land points
#
def fit_partition(partin, land):
    partout = []
    for i, dom in enumerate(partin):
        l = land[dom["jstart"]:dom["jstart"]+dom["nbj"],dom["istart"]:dom["istart"]+dom["nbi"]]
        l1 = np.sum(l, axis = 0)
        
        i0 = np.where(l1>0)[0][0]
        i1 = np.where(l1>0)[0][-1]

        l2 = np.ma.sum(l, axis = 1)
        j0 = np.where(l2>0)[0][0]
        j1 = np.where(l2>0)[0][-1]

        dom["jstart"] = j0 + dom["jstart"] 
        dom["nbj"] = j1-j0+1
        dom["istart"] = i0 + dom["istart"] 
        dom["nbi"] = i1-i0+1
        dom["nbland"] = int(np.nansum(land[dom["jstart"]:dom["jstart"]+dom["nbj"],dom["istart"]:dom["istart"]+dom["nbi"]]))

        partout.append(dom)        
    return partout
#
#
#
def adjustpart(partin, land, nbcore) :
    nbland=np.array([dom['nbland'] for dom in partin])
    nbptproc=np.array([dom['nbi']*dom['nbj'] for dom in partin])
    zeros=np.nonzero(nbland < 1)[0]
    #
    # Delete partitions without land points.
    #
    for i in range(len(zeros)) :
        partin.pop(zeros[i]-i)
    #
    # Divide domains with the largest number of land points
    #
    nbptproc=np.array([dom['nbi']*dom['nbj'] for dom in partin])
    freecores = nbcore-len(partin)

    while nbcore-len(partin) > 0:
        partin = halfpartition(partin, land)

    partin = fit_partition(partin, land)

    nbptproc=np.array([dom['nbi']*dom['nbj'] for dom in partin])
    nbland = [dom["nbland"] for dom in partin]

    if np.min(nbptproc) < 3 :
        ERROR("The smallest domain in the partition is of a size less than 5. This will probably not work.")
        sys.exit()        

    return partin
#
# Partition domain in two steps : 1) halving, 2) adjusting by nb land points
#
def partitiondom(land, nbcore, rank) :
    njg,nig = land.shape
    partout=[{"nbi":nig,"nbj":njg,"istart":0,"jstart":0, "nbland":int(np.sum(land))}]
    while len(partout) < nbcore :
        partout = halfpartition(partout, land)
    #
    partout = adjustpart(partout, land, nbcore)
    #
    gindland=np.zeros((njg,nig), dtype=np.int32)
    gindland[:,:]=-1
    test = np.zeros((njg,nig), dtype=np.int32)
    test[:,:] = land[:,:]
    n=0
    for i in range(nig) :
        for j in range(njg) :
            if (land[j,i] > 0 ) :
                gindland[j,i] = n
                n += 1
    #
    for ic in range(len(partout)) :
        istart = partout[ic]["istart"]
        nbi = partout[ic]["nbi"]
        jstart = partout[ic]["jstart"]
        nbj = partout[ic]["nbj"]
        locgind=np.copy(gindland[jstart:jstart+nbj,istart:istart+nbi])
        loccore=np.copy(gindland[jstart:jstart+nbj,istart:istart+nbi])
        loccore[loccore >= 0] = ic
        test[jstart:jstart+nbj,istart:istart+nbi] = test[jstart:jstart+nbj,istart:istart+nbi] - land[jstart:jstart+nbj,istart:istart+nbi]
        partout[ic]["glandindex"] = locgind
        partout[ic]["corenb"] = loccore
        partout[ic]["nblocland"] = np.count_nonzero(locgind >= 0)
    #
    nbland = [part["nbland"] for part in partout]
    if rank == 0:
        INFO("Distribution of the domain : {}".format(nbland))
    INFO("Test number of error on domain distribution: {0}".format(len(np.where(test == 1)[0])))
    INFO("Test domain partition (only value must be 0): {0}".format(np.unique(test)))
    #
    return partout
#
# Build global map of processors
#
def buildprocmap(nig, njg, part) :
    procmap=np.zeros((njg,nig), dtype=np.int8)
    procmap[:,:] = -1
    for i in range(len(part)) :
        for ij in range(part[i]["nbj"]) :
            for ii in range(part[i]["nbi"]) :
                procmap[ij+part[i]["jstart"],ii+part[i]["istart"]] = i
    #
    return procmap
#
# Partion of only land points in the case where we do not need halos
#
def landpartition(land, nbcore, rank) :
    nbland =  np.count_nonzero(land > 0)
    optland = int(nbland/nbcore)+(nbland%nbcore > 0)
    print("RANK",rank,"Nbland = ", np.count_nonzero(land > 0), " nbcore = ", nbcore, " optland = ",optland)
    #
    distppc=np.zeros(nbcore)
    distppc[:]=0
    for i in range(nbcore-1) :
        ptsleft = int(nbland-np.sum(distppc[0:i]))
        crsleft = nbcore-i
        if  ptsleft > optland and ptsleft/2 > crsleft :
            distppc[i] = optland
        else :
            print("On proc ", i," Still have ", ptsleft, " points to distribute and ", crsleft, " cores")
            distppc[i] = int(ptsleft/crsleft)
        #
    distppc[-1]=nbland-np.sum(distppc[0:-1])
    print("Distpcc :", distppc)
    #
    njg,nig=land.shape
    gindland=np.zeros((njg,nig), dtype=np.int32)
    gindland[:,:]=-1
    glandcore=np.zeros((njg,nig), dtype=np.int32)
    glandcore[:,:]=-1
    n=0
    for i in range(nig) :
        for j in range(njg) :
            if (land[j,i] > 0 ) :
                gindland[j,i] = n
                cc = np.argmax(distppc > 0)
                glandcore[j,i] = cc
                distppc[cc] = distppc[cc]-1
                n += 1
    #
    # Decompose domain according to land point positions
    #
    partout=[]
    for ic in range(nbcore) :
        ff = np.nonzero(glandcore == ic)
        # Region is at least 3x3
        nbi=max(max(ff[1])-min(ff[1])+1, 3)
        nbj=max(max(ff[0])-min(ff[0])+1, 3)
        istart=min(min(ff[1]),nig-nbi)
        jstart=min(min(ff[0]),njg-nbj)
        #
        loccore=np.copy(glandcore[jstart:jstart+nbj,istart:istart+nbi])
        loccore[loccore != ic] = -1
        locgind=np.copy(gindland[jstart:jstart+nbj,istart:istart+nbi])
        locgind[loccore < 0] = -1
        partout.append({"nbi":nbi, "nbj":nbj, "istart":istart, "jstart":jstart, "nbland":ff[0].shape[0], \
                        "corenb":loccore, "glandindex":locgind, "nblocland":np.count_nonzero(locgind >= 0)})
    #
    return partout
#
# Add halo to the partition
#
def addhalo(nig, njg, part, procmap, nbh) :
    nbcore=len(part)
    halosource=np.zeros((njg,nig,nbcore), dtype=np.int8)
    halosource[:,:,:] = -1
    coresend=np.zeros((njg,nig,nbcore), dtype=np.int8)
    coresend[:,:,:] = -1
    #
    for proc, dom in enumerate(part) :
        dom['ihstart']=max(dom["istart"]-nbh,0)
        dom['nbih']=min(dom["istart"]+dom['nbi']-1+nbh, nig-1)-dom['ihstart']+1
        dom['jhstart']=max(dom["jstart"]-nbh,0)
        dom['nbjh']=min(dom["jstart"]+dom['nbj']-1+nbh, njg-1)-dom['jhstart']+1
        for i in range(dom['nbih']) :
            #
            ic=dom['ihstart']+i
            #
            if dom['jhstart'] != dom["jstart"] :
                sproc = procmap[dom['jhstart'],ic]
                halosource[dom['jhstart'],ic,proc] = sproc
                coresend[dom['jhstart'],ic,proc] = proc
            #
            if dom['jhstart']+dom['nbjh']-1 != dom['jstart']+dom['nbj']-1 :
                sproc = procmap[dom['jhstart']+dom['nbjh']-1,ic]
                halosource[dom['jhstart']+dom['nbjh']-1,ic,proc] = sproc
                coresend[dom['jhstart']+dom['nbjh']-1,ic,proc] = proc
        for j in range(dom['nbjh']) :
            #
            jc=dom['jhstart']+j
            #
            if dom['ihstart'] != dom['istart'] :
                sproc = procmap[jc,dom['ihstart']]
                halosource[jc,dom['ihstart'],proc] = sproc 
                coresend[jc,dom['ihstart'],proc] = proc
            if dom['ihstart']+dom['nbih']-1 != dom['istart']+dom['nbi']-1 :
                sproc = procmap[jc,dom['ihstart']+dom['nbih']-1]
                halosource[jc,dom['ihstart']+dom['nbih']-1,proc] = sproc
                coresend[jc,dom['ihstart']+dom['nbih']-1,proc] = proc
    return halosource, coresend
#
# Get the list of the core points which need to be sent to another core.
#
def coresendlist(innersend_map, istart, ihstart, ni, jstart, jhstart, nj) :
    dcore=innersend_map[jstart:jstart+nj,istart:istart+ni,:]
    sendto = np.unique(dcore)[np.where(np.unique(dcore) >= 0)]
    innersend = []
    for ic in sendto :
        innersend.append(np.where(dcore[:,:,ic]==ic))
    #
    # Offset by halo
    #
    ihoff=istart-ihstart
    jhoff=jstart-jhstart
    for ic in range(len(sendto)) :
        innersend[ic][0][:] = innersend[ic][0][:]+jhoff
        innersend[ic][1][:] = innersend[ic][1][:]+ihoff
    return sendto, innersend
#
# All points belonging to the core of the domain
#
def listcoreland(istart, ihstart, ni, jstart, jhstart, nj, landind) :
    corelandlist=[]
    ihoff=istart-ihstart
    jhoff=jstart-jhstart
    for i in range(ni) :
        for j in range(nj) :
            if landind[jhoff+j,ihoff+i] >= 0 :
                corelandlist.append(landind[jhoff+j,ihoff+i])
    return corelandlist
#
# Get list of halo points which need to receive data
#
def haloreceivelist(halosource_map, rank) :
    receivefrom = np.unique(halosource_map[:,:,rank])[np.where(np.unique(halosource_map[:,:,rank]) >= 0)]
    halosource_g = []
    for ic in receivefrom :
        halosource_g.append(np.where(halosource_map[:,:,rank]==ic))
    return receivefrom, halosource_g
#
# Get 1D indices for the land points 
#
def landindexmap(istart, ni, jstart, nj, land) :
    gnj,gni=land.shape
    gindland=np.zeros((gnj,gni), dtype=np.int32)
    gindland[:,:]=-1
    n=0
    for i in range(gni) :
        for j in range(gnj) :
            if (land[j,i] > 0 ) :
                gindland[j,i] = n
                n += 1
    #
    indland=np.zeros((nj,ni), dtype=np.int32)
    indland[:,:]=-1
    local2global=[]
    n=0
    for i in range(ni) :
        for j in range(nj) :
            if (land[jstart+j,istart+i] > 0 ) :
                indland[j,i] = n
                local2global.append(gindland[jstart+j,istart+i])
                n += 1
    return n, indland, np.array(local2global, dtype=np.int32)
#
# Convert indices to local
#
def tolocal_index(x, istart, jstart) :
    xout = []
    for c in x :
        xout.append([c[0]-jstart,c[1]-istart])
    return xout
#
#
#
def toland_index(x,landmap) :
    xout=[]
    for c in x :
        vl = landmap[c[0],c[1]]
        if np.count_nonzero(vl >= 0) > 0 :
            xout.append(vl[np.where(vl >= 0)])
        else :
            xout.append([])
    return xout
#
# 
#
class partition :
    def __init__ (self, nig, njg, land, mpicomm, nbcore, halosz, rank, wunit="None") :
        #
        # Self varibales
        #
        # MPI info
        self.comm = mpicomm
        self.size = nbcore
        self.rank = rank
        # Global info
        self.nig = nig
        self.njg = njg
        self.nblandg = np.count_nonzero(land > 0)
        #
        part = partitiondom(land, nbcore, rank)
        procmap = buildprocmap(nig, njg, part)
        # To be activated when halosz = 0 and we can deal with single point domain grids.
        #
        #     part = landpartition(land, nbcore, rank)
        #     procmap = buildprocmap(nig, njg, part)
        #
        halosource_map, innersend_map = addhalo(nig, njg, part, procmap, halosz)
        #
        INFO(" Nb of land points on proc "+str(rank)+" = "+str(part[rank]["nblocland"])+" from "+str(self.nblandg)+\
         " -- "+"%.2f"%(part[rank]["nblocland"]/self.nblandg*100)+"% of total.")
        #
        self.allistart = []
        self.alljstart = []
        for i in range(self.size) :
            self.allistart.append(part[i]["istart"])
            self.alljstart.append(part[i]["jstart"])
        #
        # Information of domain on local processor
        #
        self.ni = part[rank]["nbi"]
        self.nj = part[rank]["nbj"]
        self.istart = part[rank]["istart"]
        self.jstart = part[rank]["jstart"]
        self.nih = part[rank]["nbih"]
        self.njh = part[rank]["nbjh"]
        self.ihstart = part[rank]["ihstart"]
        self.jhstart = part[rank]["jhstart"]
        #
        self.nbland, landindmap, self.l2glandind = landindexmap(self.ihstart, self.nih, self.jhstart, self.njh, land)        
        #
        if wunit != "None" :
            wunit.write("Offsets with halo (j-i) : "+str(self.jhstart)+"-"+str(self.ihstart)+'\n')
            wunit.write("Offsets without halo (j-i) : "+str(self.jstart)+"-"+str(self.istart)+'\n')
            wunit.write("Sizes without halo (j-i) : "+str(self.nj)+"-"+str(self.ni)+'\n')
            wunit.write("Shape of halosource_map"+str(halosource_map.shape)+"\n")
            for j in range(self.njh) :
                wunit.write(str(halosource_map[self.jhstart+j,self.ihstart:self.ihstart+self.nih,rank])+"\n")
            wunit.write("=============\n")
            wunit.write("Shape of innersend_map"+str(innersend_map.shape)+" from proc "+str(rank)+"\n")
            for r in range(self.size) :
                smtmp = innersend_map[self.jstart:self.jstart+self.nj,self.istart:self.istart+self.ni,r]
                if np.sum(smtmp) > -1.0*self.nj*self.ni :
                    wunit.write(str(r)+" ---------------- "+str(r)+"\n")
                    for j in range(self.nj) :
                        wunit.write(str(smtmp[j,:])+"\n")
            wunit.write("=============\n")
            wunit.write("Shape of landinmap :"+str(landindmap.shape)+"\n")
            wunit.write(str(landindmap)+"\n")
            wunit.write("=============\n")
        #
        self.receivefrom, self.halosource_g = haloreceivelist(halosource_map, rank)
        self.nbreceive = len(self.receivefrom)
        # To local indexing
        self.halosource = tolocal_index(self.halosource_g, self.ihstart, self.jhstart)
        self.landhalosrc = toland_index(self.halosource, landindmap)
        #
        if wunit != "None" :
            wunit.write("+++++ Halo receivefrom list : "+str(self.receivefrom)+"\n")
            for ic in range(self.nbreceive) :
                wunit.write("For "+str(self.receivefrom[ic])+" landhalosrc is :"+str(self.landhalosrc[ic])+"\n")
        #
        self.sendto,self.innersend = coresendlist(innersend_map, self.istart, self.ihstart, self.ni, self.jstart, self.jhstart, self.nj)
        self.nbsend = len(self.sendto)
        # To global indexing
        self.innersend_g = self.toglobal_index(self.innersend)
        self.landinnersend = toland_index(self.innersend, landindmap)
        if wunit != "None" :
            wunit.write("+++++ Halo core send list : "+str(self.receivefrom)+"\n")
            wunit.write("+++++ Core send list : "+str(self.sendto)+"\n")
            for ic in range(self.nbsend) :
                wunit.write("To "+str(self.sendto[ic])+" send innersend :"+str(self.innersend[ic])+"\n")
                wunit.write("To "+str(self.sendto[ic])+" send landinnersend :"+str(self.landinnersend[ic])+"\n")
        #
        self.landcorelist = listcoreland(self.istart, self.ihstart, self.ni, self.jstart, self.jhstart, self.nj, landindmap)
        #
        return
    #
    # Convert indices from local to global
    #
    def toglobal_index(self, x) :
        xout = []
        for c in x :
            xout.append([c[0]+self.jhstart,c[1]+self.ihstart])
        return xout
    #
    # Send halo to the other procs
    #
    def sendtohalo(self, x) :
        #
        for i in range(max(self.nbreceive,self.nbsend)) :
            if i < self.nbsend :
                if len(x.shape) == 2 :
                    self.comm.send(x[self.innersend[i][0],self.innersend[i][1]], dest=self.sendto[i])
                else :
                    ERROR("Unforessen rank of the variable to be sent to halo")
                    sys.exit()
            if i < self.nbreceive :
                if len(x.shape) == 2 :
                    x[self.halosource[i][0],self.halosource[i][1]] = self.comm.recv(source=self.receivefrom[i])
                else :
                    ERROR("Unforessen rank of the variable to be received in halo")
                    sys.exit()
        return
    #
    # For field gathered in land point send halo to the other procs
    #
    def landsendtohalo(self, x, order='C') :
        #
        # A simple vector
        #
        if len(x.shape) == 1 :
            for i in range(max(self.nbreceive,self.nbsend)) :
                if i < self.nbsend :
                    if len(self.landinnersend[i]) > 0 :
                        self.comm.send(x[self.landinnersend[i]], dest=self.sendto[i])
                if i < self.nbreceive :
                    if len(self.landhalosrc[i]) > 0 :
                        x[self.landhalosrc[i]] = self.comm.recv(source=self.receivefrom[i])
        #
        # Working on matrices
        #
        elif len(x.shape) == 2 :   
            chksz = 100
            if order == 'C' :
                chkdim = x.shape[0]
            elif order == 'F' :
                chkdim = x.shape[1]
            else :
                ERROR("Unforessen order of the variable to be sent to halo of land points")
                sys.exit()
            chks = list(range(0, chkdim, chksz))
            chks.append(chkdim)
            #
            for j in range(len(chks)-1) :
                for i in range(max(self.nbreceive,self.nbsend)) :
                    if i < self.nbsend :
                        if len(self.landinnersend[i]) > 0 :
                            if order == 'C' :
                                self.comm.send(x[chks[j]:chks[j+1],self.landinnersend[i]], dest=self.sendto[i])
                            elif order == 'F' :
                                self.comm.send(x[self.landinnersend[i],chks[j]:chks[j+1]], dest=self.sendto[i])
                    if i < self.nbreceive :
                        if len(self.landhalosrc[i]) > 0 :
                            if order == 'C' :
                                x[chks[j]:chks[j+1],self.landhalosrc[i]] = self.comm.recv(source=self.receivefrom[i])
                            elif order == 'F' :
                                x[self.landhalosrc[i],chks[j]:chks[j+1]] = self.comm.recv(source=self.receivefrom[i])
        #
        # Some higher ranked variable
        #
        else :
            ERROR("Unforessen rank of the variable to be received in halo of land points")
            sys.exit()
     
        return
    #
    # Gather all fields partitioned in the 2D domain onto the root proc
    #
    def gather(self, x) :
        #
        indim = x.shape
        outdim = indim[:-2]+(self.njg,self.nig)
        if self.rank == 0 :
            xout = np.zeros(outdim, dtype=x.dtype)
            if len(outdim) == 2 :
                xout[:,:] = np.nan
            elif len(outdim) == 3 :
                xout[:,:,:] = np.nan
            elif len(outdim) == 4 :
                xout[:,:,:,:] = np.nan
            else :
                ERROR("Unforessen rank of field to be gathered")
                sys.exit()
                
        else :
            xout = None
        #
	# Offset for the core region
	#
        ihoff=self.istart-self.ihstart
        jhoff=self.jstart-self.jhstart
        #
        if len(outdim) == 2 :
            xtmp = self.comm.gather(x[jhoff:jhoff+self.nj,ihoff:ihoff+self.ni], root=0)
        elif len(outdim) == 3 :
            xtmp = self.comm.gather(x[:,jhoff:jhoff+self.nj,ihoff:ihoff+self.ni], root=0)
        elif len(outdim) == 4 :
            xtmp = self.comm.gather(x[:,:,jhoff:jhoff+self.nj,ihoff:ihoff+self.ni], root=0)
        else :
            ERROR("Unforessen rank of field to be gathered")
            sys.exit()
        #
        if self.rank == 0 :
            for i,z in enumerate(xtmp) :
                zdim=z.shape
                if len(outdim) == 2 :
                    xout[self.alljstart[i]:self.alljstart[i]+zdim[-2],self.allistart[i]:self.allistart[i]+zdim[-1]] = z[:,:]
                elif len(outdim) == 3 :
                    xout[:,self.alljstart[i]:self.alljstart[i]+zdim[-2],self.allistart[i]:self.allistart[i]+zdim[-1]] = z[:,:,:]
                elif len(outdim) == 4 :
                    xout[:,:,self.alljstart[i]:self.alljstart[i]+zdim[-2],self.allistart[i]:self.allistart[i]+zdim[-1]] = z[:,:,:,:]
                else :
                    ERROR("Unforessen rank of field to be gathered")
                    sys.exit()
        #
        return xout
    #
    # For land vectors gathered over land and local to the proc, we gather them on root proc.
    #
    def landscatgat(self, modelgrid, xl, order='C') :
        xf = modelgrid.landscatter(xl, order=order)
        xout = self.gather(xf)
        return xout
    #
    # Convert local index of land points to global index
    #
    def l2glandindex(self, x) :
        if x.ndim == 2:
           nl,nh = x.shape
           y = np.zeros(x.shape, dtype=x.dtype)
           if nl == self.nbland :
               for i in range(nl) :
                   for j in range(nh) :
                       # Land indices are in FORTRAN !!
                       y[i,j] = self.l2glandind[x[i,j]-1]+1
           else :
               ERROR("The first dimension does not have the length of the number of land points")
               sys.exit()
        if x.ndim == 3:
           nl,nh1,nh2 = x.shape
           y = np.zeros(x.shape,dtype = x.dtype)
           if nl == self.nbland:
               for i in range(nl):
                   for j in range(nh1):
                       for k in range(nh2):
                           # Land indices are in Fortran
                           y[i,j,k] = self.l2glandind[x[i,j,k]-1]+1
           else:
               ERROR("The first dimension does not have the length of the number of land points")
               sys.exit()
        return y
    #
    # Set to zero all points in the core
    #
    def zerocore(self, x, order='C') :
        if order == 'C' :
            if x.shape[-1] != self.nbland :
                ERROR("Space dimension does not correspond to number of land points (case C)")
                sys.exit()
        elif order == 'F' :
            if x.shape[0] != self.nbland :
                ERROR("Space dimension does not correspond to number of land points (case F)")
                sys.exit()
        else :
            ERROR("Unforessen order for variable to be set to zero over core region.")
            sys.exit()
        #
        y = np.copy(x)
        #
        if len(x.shape) == 1 :
            y[self.landcorelist] = 0.0
        elif len(x.shape) == 2 :
            if order == 'C' :
                y[:,self.landcorelist] = 0.0
            elif order == 'F' :
                y[self.landcorelist,:] = 0.0
        else :
            ERROR("Unforessen rank for variable to summedibe set to zero over the core region.")
            sys.exit()
        return y
    #
    # Copy all points in the core of y into x to produce z
    #
    def copycore(self, x, y, order='C') :
        if order == 'C' :
            if x.shape[-1] != self.nbland :
                ERROR("Space dimension does not correspond to number of land points (case C)")
                sys.exit()
        elif order == 'F' :
            if x.shape[0] != self.nbland :
                ERROR("Space dimension does not correspond to number of land points (case F)")
                sys.exit()
        else :
            ERROR("Unforessen order for variable to be set to zero over core region.")
            sys.exit()
        #
        z = np.copy(x)
        #
        if len(x.shape) == 1 :
            z[self.landcorelist] = y[self.landcorelist]
        elif len(x.shape) == 2 :
            if order == 'C' :
                z[:,self.landcorelist] = y[:,self.landcorelist]
            elif order == 'F' :
                z[self.landcorelist,:] = y[self.landcorelist,:]
        else :
            ERROR("Unforessen rank for variable to summedibe set to zero over the core region.")
            sys.exit()
        return z
    #
    # Function to sum over the core regions
    #
    def sumcore(self, x, order='C') :
        if order == 'C' :
            if x.shape[-1] != self.nbland :
                ERROR("Space dimension does not correspond to number of land points (case C)")
                sys.exit()
        elif order == 'F' :
            if x.shape[0] != self.nbland :
                ERROR("Space dimension does not correspond to number of land points (case F)")
                sys.exit()
        else :
            ERROR("Unforessen order for variable to summed over the core region.")
            sys.exit()
        #
        if len(x.shape) == 1 :
            y = np.nansum(x[self.landcorelist])
        elif len(x.shape) == 2 :
            if order == 'C' :
               y =  np.nansum(x[:,self.landcorelist], axis=1)
            elif order == 'F' :
               y = np.nansum(x[self.landcorelist,:], axis=0)
        else :
            ERROR("Unforessen rank for variable to summed over the core region.")
            sys.exit()
        return y
    #
    # Function to find max value over the entire domain
    #
    def domainmax(self, x) :
        return max(self.comm.allgather(x))
    #
    # Function to gather vectors of points on different procs on root
    #
    def gatherbypoint(self, x) :
        maxlen = self.domainmax(len(x))
        if self.rank == 0 :
            xout = np.empty((maxlen,self.size))
            # Receive from all procs
            for lr in range(self.size) :
                if lr == 0 :
                    xout[:,lr] = np.pad(x, [0,maxlen-len(x)], mode='constant', constant_values=np.nan)
                else :
                    xout[:,lr] = self.comm.recv(source=lr)
        else :
            xout = None
            self.comm.send(np.pad(x, [0,maxlen-len(x)], mode='constant', constant_values=np.nan), dest=0)
        #
        return xout
    
