#
#
#
import numpy as np
import os, sys
from mpi4py import MPI
import gc
#
import matplotlib as mpl
mpl.use('Agg')
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import time
#
# Gert the information from the configuration file.
#
import configparser
config=configparser.ConfigParser()
config.read("run.def")
nbasmax=config.getint("OverAll", "nbasmax")
numop=config.getint("OverAll", "numop", fallback=100)
OutGraphFile=config.get("OverAll","GraphFile")
DumpHydroSuper=config.getboolean("OverAll","DumpHydroSuper",fallback=False)
wfile=config.get("OverAll","WeightFile",fallback="Weights.nc")
#
import ModelGrid as MG
import Partition as PA
import RPPtools as RPP
import HydroGrid as HG
import Interface as IF
from Truncate import trunc as TR
#
# Logging in MPI : https://groups.google.com/forum/#!topic/mpi4py/SaNzc8bdj6U
#
import getargs
log_master, log_world = getargs.getLogger()
INFO, DEBUG, ERROR = log_master.info, log_master.debug, log_world.error
INFO_ALL, DEBUG_ALL = log_world.info, log_world.debug
#
# Read full grid and partition the domain.
#
gg = MG.GlobalGrid()
#
comm = MPI.COMM_WORLD
szhalo=1
nbcore=comm.Get_size()
rank=comm.Get_rank()
#
# Region of grid to be treated
#
part = PA.partition(gg.ni, gg.nj, gg.land, comm, nbcore, szhalo, rank)
#
modelgrid=MG.ModelGrid(part.ihstart+gg.igstart,part.nih,part.jhstart+gg.jgstart,part.njh)
#
INFO("Longitude interval on proc "+str(rank)+" = "+str(modelgrid.box_land[0][0])+" : "+str(modelgrid.box_land[0][1]))
INFO("Latitude interval on proc "+str(rank)+" = "+str(modelgrid.box_land[1][0])+" : "+str(modelgrid.box_land[1][1]))
#
hydrogrid=HG.HydroGrid(modelgrid.box_land, wfile)
#
# Computes weights of overlap of modelgrid and hydrogrid
w = RPP.compweights(wfile, part, modelgrid, hydrogrid)
#
#
#
nbpt = len(w.index)
nbvmax = part.domainmax(max(w.hpts))
INFO("nbpt : {0}".format(nbpt))
INFO("nbvmax : {0}".format(nbvmax))
INFO("nbasmax : {0}".format(nbasmax))
#
# Extract hydo data from file
#
INFO("=================== HYDRODATA ====================")
hydrodata = HG.HydroData(hydrogrid.ncfile, hydrogrid.box, w.index)

INFO("=================== INITATMGRID ====================")
IF.initatmgrid(rank, nbcore, nbpt, modelgrid)

INFO("=================== HOVERLAP ====================")
hoverlap = IF.HydroOverlap(nbpt, nbvmax, w.hpts, w.index, w.area, w.lon, w.lat, part, modelgrid, hydrodata)
#
# Do some memory management and synchronize procs.
#

del w
gc.collect()
comm.Barrier()

hsuper = IF.HydroSuper(nbvmax, hydrodata, hoverlap, nbasmax, part)
#
INFO("=================== LINKUP ====================")
hsuper.linkup(hydrodata)

#
# Do some memory management and synchronize procs.
#

del hoverlap
gc.collect() 
comm.Barrier()

INFO("=================== Compute fetch ====================")
t = time.time()
hsuper.fetch(part)
comm.Barrier()
t1 = time.time()
print("Time for fetch: {:0.2f} s.".format(t1-t)) 
comm.Barrier()

if DumpHydroSuper :
    INFO("Dumping HydroSuper")
    hsuper.dumpnetcdf(OutGraphFile.replace(".nc","_HydroSuper.nc"), gg, modelgrid, part)

hsuper.check_inflows()

INFO("=================== Truncate ====================")
print("Rank : {0} - Basin_count Before Truncate : ".format(part.rank), hsuper.basin_count)
hs = TR(hsuper, part, comm, modelgrid, numop = numop)
print("Rank : {0} - Basin_count After Truncate : ".format(part.rank), hsuper.basin_count)

INFO("=================== HydroGraph ====================")
hgraph = IF.HydroGraph(nbasmax, hsuper, part, modelgrid)
hgraph.dumpnetcdf(OutGraphFile, gg, modelgrid, part)

IF.closeinterface(comm)
