import numpy as np
import datetime
import math
import sys
#
from difflib import SequenceMatcher
#
# For Json format get dates
#
def key2date(obj, key) :
    dates = []
    for dd in obj[key] :
        dates.append(datetime.datetime(int(dd[0:4]), int(dd[4:6]), int(dd[6:8])))
    return dates
#
# For Json format get data
#
def key2data(obj, key) :
    nval=len(obj[key].keys())
    data=np.zeros((nval))
    for i,dd in enumerate(obj[key]) :
        data[i] = obj[key][dd]
    return data
#
# Compute montly means when original data is daily
#
def monthlymean(d,x) :
    inyrs = np.array([di.year for di in d])
    inym = np.array([(di.year*100)+di.month for di in d])
    # Carefull data can have gaps, We recreate a list of years
    nbinyrs = inyrs[-1]-inyrs[0]+1
    nd = np.arange(nbinyrs)+inyrs[0]
    # New data
    od = []
    ox = []
    #
    for iy in range(nbinyrs) :
        for im in range(12) :
            ind = (inym==nd[iy]*100+(im+1))
            mday = datetime.datetime(nd[iy], im+1, 15)
            mlen = (mday.replace(month = mday.month % 12 +1, day = 1)-datetime.timedelta(days=1)).day
            od.append(mday)
            if len(x[ind]) < mlen :
                # See if we can fix that
                ox.append(np.nan)
            else :
                ox.append(np.mean(x[ind]))
            #
    return od, np.array(ox)
#
# get time axis from a NetCDF filex
#
def gettimeaxis(nf):
    dims=nf.dimensions.keys()
    for d in dims:
        if ( d.find('time') >= 0 ):
            tname=d
    t=np.copy(nf.variables[tname][:])
    tunits=getattr(nf.variables[tname],"units")
    if ( tunits.find("seconds") >= 0):
        epoch=datetime.datetime.strptime(tunits[tunits.find("since ")+6:], "%Y-%m-%d %H:%M:%S")
        date_list = [epoch+datetime.timedelta(seconds=x.astype(np.float64)) for x in t]
    elif( tunits.find("days") >= 0):
        epoch=datetime.datetime.strptime(tunits[tunits.find("since ")+6:], "%Y-%m-%d %H:%M:%S")
        date_list = [epoch+datetime.timedelta(days=x.astype(np.float64)) for x in t]
    elif( tunits.find("month") >= 0):
        epoch=datetime.datetime.strptime(tunits[tunits.find("since ")+6:], "%Y-%m-%d %H:%M:%S")
        date_list=[]
        for x in t :
            m=((epoch.month+(x-1))%12)+1
            y=epoch.year+int((x-m)/12)+1
            date_list.append(datetime.date(int(y),int(m),epoch.day))
    else :
        print("The time axis is not in seconds and I cannot yet deal with it")
        sys.exit()
    return date_list
#
#
#
def distance_haversine(A, B, radius=6371000.):
    """ note that the default distance is in meters """
    dLat = math.radians(B[1]-A[1])
    dLon = math.radians(B[0]-A[0])
    lat1 = math.radians(A[1])
    lat2 = math.radians(B[1])
    a = math.sin(dLat/2) * math.sin(dLat/2) + math.sin(dLon/2) * math.sin(dLon/2) * math.cos(lat1) * math.cos(lat2)
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1-a))
    return c * radius
#
# Function to score strings
#
def scorestring(stra, strb) :
    # One string is in another
    match = SequenceMatcher(None,stra,strb).find_longest_match(0, len(stra), 0, len(strb))
    #
    if match.size == min(len(stra), len(strb)) :
        # Sub-string case
        score = 1.0
    else :
        # Fuzzy check
        score = SequenceMatcher(None,stra,strb).ratio()
    #
    return score
#
# Function to put data onto a new time axis
#
def tonewtime(t,x,newt) :
    newx = np.zeros((len(newt)))
    newx[:] = np.nan
    intersect=np.intersect1d(t,newt)
    if len(intersect) == 0 :
        print("Input time goes from ", t[0]," to ", t[-1])
        print("Target time axis goes from ",newt[0], " to ", newt[-1])
        print("Intersection is empty !")
        sys.exit()
    newx[newt.index(intersect[0]):newt.index(intersect[-1])]=x[t.index(intersect[0]):t.index(intersect[-1])]
    return newx
#
# Merge a number of data set which are already on the same time axis
#
def mergedata(d, Merged, verbose=False) :
    # Options
    # Merger quality criteria
    q = 0.1
    # Method to fill gaps : First or Mean
    opt = "Mean"
    #
    da=np.reshape(np.append(Merged,np.asarray(d)),(-1,Merged.shape[0]))
    nbdata,tlen=da.shape
    nbval=np.zeros((nbdata))
    overlapmean=np.zeros((nbdata))
    mask=~np.isnan(np.sum(da,axis=0))
    grandmean=np.nanmean(da)
    #
    for i in range(nbdata) :
        nbval[i] = np.count_nonzero(~np.isnan(da[i,:]))
        overlapmean[i] = np.mean(da[i,mask])
        if verbose :
            print(" mergedata : Mean of obs. = ", np.nanmean(da[i,:]), " Nb obs = ", nbval[i], \
                  " Mean of overlap period = ", np.mean(da[i,mask]))
    #
    # Default us the Merged product from the selected station
    #
    c=np.copy(da[0,:])
    #
    # If the various observations are close to each other we have two options to fill the gaps :
    # Take the first product which is available, or use the main of all products
    #
    if np.std(overlapmean)/np.mean(overlapmean) < q :
        if opt.find("Mean") != -1 :
            # A mask without the Merged data
            mask=~np.isnan(np.sum(da[1:,:],axis=0))
            m = np.zeros((tlen))
            m[:] = np.nan
            m[mask] = np.nanmean(da[:,mask], axis=0)
            c[np.isnan(c)]=m[np.isnan(c)]
        # Fill with the other data when only option is available
        for i in range(nbdata) :
            c[np.isnan(c)]=da[i,np.isnan(c)]
    #
    return c

