#
# Class to store individual stations with all the metadata.
# The observations are put on a continuous time axis from the
# date of the first observation to the last. Missing observations
# are noted as np.nan.
#
#
import sys
import numpy as np
import datetime
import re
from netCDF4 import stringtochar, chartostring
#
import ConversionCalculations as CC
#
defint = -999
defstr = "XXX"
#
class Station :
    def __init__(self, ftype, filename, index=0, regioncode=-999, countrycode="XX") :
        #
        # Set default values for the station
        #
        self.sourceproj = None
        self.lonlatproj = None
        #
        # GRDC station number
        self.Number = defint
        self.WMOregion = defint
        self.WMOsubregion = defint
        self.NextStation = defint
        self.River = defstr
        self.Name = defstr
        self.FileDate = defstr
        self.LastUpdate = defstr
        self.Country = defstr
        self.Area = defint
        self.Altitude = defint
        self.Length = defint
        self.Location = None
        self.Time = None
        self.Original = None
        self.mergersource = None
        self.Merged = None
        self.Flag = None
        #
        # 
        #
        if ftype.find("GRDCascii") >= 0 :
            self.GRDCasciiread(filename)
        elif ftype.find("SpainJson") >= 0 :
            self.Spainjsonread(filename, countrycode, regioncode)
        elif ftype.find("netcdf") >= 0 :
            self.getfromnetcdf(filename, index)
        else:
            print("Not foreseen file format :", ftype)
            sys.exit()
        #
        #
        #
    def GRDCasciiread(self, filename) :
        #
        f = open(filename, 'rt', encoding='iso8859_9')
        #
        skip=0
        nbdata=-1
        for line in f:
            #
            # Find position of #If "#" is in first position print line
            #
            if ( line.find("#") == 0 ) :
                if ( line.find("GRDC-No.:") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.Number=int(tmpread)
                    self.WMOregion = (self.Number/100000)
                    self.WMOsubregion = int((self.Number-self.WMOregion*100000)/1000)
                elif ( line.find("Next downstream station:") > 0 ) :
                    start=line.find("Next downstream station:")+len("Next downstream station:")+1
                    tmpread=line[start:]
                    if len(tmpread.strip()) <= 1 or tmpread.find('-') >= 0 :
                        self.NextStation = -999
                    else :
                        self.NextStation=int(tmpread)
                elif (line.find("River:") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.River=tmpread.strip().title()
                elif (line.find("file generation") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.FileDate=tmpread.strip()
                elif (line.find("Last update") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.LastUpdate=tmpread.strip()
                elif (line.find("Station:") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:].lower()
                    self.Name=tmpread.strip().title()
                elif (line.find("Country:") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.Country=tmpread.strip()
                elif (line.find("Latitude (") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    lat=float(tmpread)
                elif (line.find("Longitude (") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    lon=float(tmpread)
                elif (line.find("Catchment area (") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.Area=float(tmpread)
                elif (line.find("Altitude (m ASL)") > 0 ) :
                    start=line.find(":")+1
                    tmpread=line[start:]
                    self.Altitude=float(tmpread)
                else :
                    # Skip line !
                    skip=skip+1
            else :
                if ( skip > 0 and nbdata < 0) :
                    if (line.find("YYYY-MM-DD;hh:mm; Original; Calculated; Flag") != 0 ) :
                        print("Unknown format, cannot read !")
                        print(line)
                        sys.exit()
                    else :
                        nbdata=0
                        date=[]
                        orig=[]
                        calc=[]
                        flag=[]
                else :
                    nbdata=nbdata+1
                    pos=[0]
                    for m in re.finditer(';', line):
                        pos.append(m.start())
                    if ( len(pos) == 5 ) :
                        day=line[pos[0]:pos[1]]
                        time=line[pos[1]+1:pos[2]]
                        original=float(line[pos[2]+1:pos[3]])
                        calculated=float(line[pos[3]+1:pos[4]])
                        dataflag=int(line[pos[4]+1:-1])
                    else :
                        print("We have not found the needed 4 ';'", nbdata)
                        print(pos)
                        print(line)
                        sys.exit()
                    y=int(day[0:4])
                    m=int(day[5:7])
                    d=15
                    date.append(datetime.datetime(y, m, d))
                    orig.append(original)
                    calc.append(calculated)
                    flag.append(dataflag)
                    #
             # Closing IF on comments
        # End of loop through lines
        f.close
        #
        # Loop through all values to build a merged product between the
        # the original month values and the montly values calculated by GRDC
        # based on daily observations
        #
        orig=np.asarray(orig, float)
        calc=np.asarray(calc, float)
        merged=np.empty((nbdata), float)
        for i in range(nbdata):
            if ( flag[i] > 80 ) and ( flag[i] <= 100 ) :
                merged[i] = calc[i]
            elif (flag[i] <= 80) and  (flag[i] >= 0):
                merged[i] = orig[i]
            else :
                print("In building the merged product a flag was not recognized")
                print("Working on ", filename)
                print("Flag = ", flag[i])
                print("orig[i] = ", orig[i], "calc[i] = ", calc[i])
                sys.exit()
        #
        # Clean-up !
        #
        self.Length=nbdata
        self.Location=[lon,lat]
        self.Time=date
        #
        orig[orig <= -999] = np.nan
        self.Original=[np.asarray(orig)]
        self.mergersource = ["Original monthly Hydrographs"]
        #
        calc[calc <= -999] = np.nan
        if np.count_nonzero(~np.isnan(calc)) > 0 :
            self.mergersource.append("Computed at GRDC from daily values")
            self.Original.append(calc)

        self.Merged=merged
        self.Merged[self.Merged <= -999]=np.nan
        self.Flag=np.asarray(flag, float)
        #
        return
        #
        #
        #
    def Spainjsonread(self, filename, countrycode, regioncode):
        #
        import json
        from pyproj import Proj, transform
        #
        if self.sourceproj is None :
            self.sourceproj = Proj(init="EPSG:25830")
        if self.lonlatproj is None :
            self.lonlatproj = Proj(init = 'epsg:4326')
        #
        f = open(filename, 'r')
        data = f.read()
        fobj = json.loads(data)
        #
        if "runoff" in fobj :
            if len(fobj["runoff"].keys()) > 0 :
                # Get coordinates
                x1,y1=float(fobj["metadata"][" UTM X H30 ETRS89"])*1000,float(fobj["metadata"][" UTM Y H30 ETRS89"])*1000
                lon,lat = transform(self.sourceproj,self.lonlatproj,x1,y1)
                #
                if " RIO" in fobj["metadata"] :
                    self.River=fobj["metadata"][" RIO"].title().strip()
                else :
                    self.River=fobj["metadata"][" Confed. Hidrografica"].title()
                self.Name=fobj["metadata"][" Municipio"].title().strip()
                self.Country=countrycode
                self.WMOregion=regioncode
                self.WMOsubregion=99
                if " Superficie aguas arriba (km2)" in fobj["metadata"] :
                    if len(fobj["metadata"][" Superficie aguas arriba (km2)"]) > 1 :
                        ## Why this 1000 ? Units and numbers dont match !
                        self.Area=float(fobj["metadata"][" Superficie aguas arriba (km2)"])*1000.0
                if " Altitud (m)" in fobj["metadata"] :
                    if len(fobj["metadata"][" Altitud (m)"]) > 1 :
                        self.Altitude=float(fobj["metadata"][" Altitud (m)"])
                self.Location=[lon,lat]
                #
                # Get data and build monthly averages
                #
                dates = CC.key2date(fobj, "runoff")
                runoff = CC.key2data(fobj, "runoff")
                month, monthlyrunoff = CC.monthlymean(dates, runoff)
                #
                self.Length=len(month)
                self.Time=month
                self.Original=monthlyrunoff
                self.Merged=monthlyrunoff
                #
            #
            #
            #
            else :
                print("Station in "+filename+" has zero runoff values")
                #
        elif "volume" in fobj :
            print("Dam named "+fobj["metadata"]["Embalse"]+" at "+fobj["metadata"][" Municipio"])
        else :
            print("Unrecognized file ",filename)
            print("Unrecognized file : Available keys : ", fobj.keys())
        #
        f.close()
        #
        return

    def getfromnetcdf(self, nc, i):        
        #
        self.Number = nc.variables["number"][i]
        self.WMOregion = nc.variables["WMOreg"][i]
        self.WMOsubregion = nc.variables["WMOsubreg"][i]
        self.NextStation = nc.variables["next"][i]
        self.River = chartostring(nc.variables["river"][i,:])
        self.Name = chartostring(nc.variables["name"][i,:])
        self.FileDate = chartostring(nc.variables["FileDate"][i,:])
        self.LastUpdate = chartostring(nc.variables["LastUpdate"][i,:])
        self.Country = chartostring(nc.variables["country"][i,:])
        self.Area = nc.variables["area"][i]
        self.Altitude = nc.variables["altitude"][i]
        self.Location = np.append(np.copy(nc.variables["lon"][i]),np.copy(nc.variables["lat"][i]))
        # Convert time to 
        self.Time = CC.gettimeaxis(nc)
        #
        FillOrig = nc.variables["hydrographs"]._FillValue
        if "alternative_index" in nc.variables.keys() :
            # Get names of alternatives
            nbalt,nbstations = nc.variables["alternative_index"].shape
            tmpmerge = [defstr]*(nbalt+1)
            for (a,v) in nc.__dict__.items():
                if a.find("MergeSource") == 0 :
                    ind = int(a[-1])-1
                    tmpmerge[ind]=v
            # Get indices
            self.mergersource = [tmpmerge[0]]
            self.Original = [nc.variables["hydrographs"][:,i]]
            for ia in range(nbalt) :
                if nc.variables["alternative_index"][ia,i] >= 0 :
                    self.mergersource.append(tmpmerge[ia+1])
                    self.Original.append(nc.variables["alt_hydrograph"][:,nc.variables["alternative_index"][ia,i]])
        else :
            self.mergersource = [nc.history]
            self.Original = [nc.variables["hydrographs"][:,i]]
        #
        #
        for d in self.Original :
            d[d >= FillOrig] = np.nan
        #
        if "mergedhydro" in nc.variables.keys() :
            self.Merged = nc.variables["mergedhydro"][:,i]
            FillMerge = nc.variables["mergedhydro"]._FillValue
            self.Merged[self.Merged >= FillMerge] = np.nan
        if "flags" in nc.variables.keys() :
            self.Flag = nc.variables["flags"][:,i]
        #
        self.Length=len(self.Time)
        #
        return
