#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 22 10:19:21 2020

@author: anthony
"""

from netCDF4 import Dataset as NetCDFFile
import numpy as np
import numpy.ma as ma
import time
import sys
import os
from inspect import currentframe, getframeinfo
localdir=os.path.dirname(getframeinfo(currentframe()).filename)
sys.path.append(localdir+'/F90subroutines')
import routing_interface
import getargs
log_master, log_world = getargs.getLogger()
INFO, DEBUG, ERROR = log_master.info, log_master.debug, log_world.error

######################

class trunc:
    """
    __init__ : realize all the truncate using the methods from the class

    This class is organized in 3 type of methods
    1. METHOD FOR EACH STEP
    Each of the 5 steps has its method that organize the (tokill, totakeover)
    2. METHOD to GET THE LIST OF CHANGES
    When needed (for step 1 to 4), these are the method to find the list of changes
    3. BASIC METHODS
    Basic methods used
    """

    def __init__(self, hydrosuper, part, comm,modelgrid, numop = 100):
        """
        __init__ : is the main program, it realize all the truncate
        """
        #
        # Initialize some variables
        # 
        self.nbpt = int(hydrosuper.nbpt)
        self.nbasmax = hydrosuper.nbasmax
        self.gridarea = modelgrid.area
        self.landcorelist = part.landcorelist
        self.landcorelist_f = [l+1 for l in part.landcorelist] 
        self.landhalolist_f=[l+1 for l in range(hydrosuper.nbpt) if l not in  self.landcorelist]
        self.numop = numop

        # Step 1 : coastal flow
        self.max_ops_num = numop
        i = 1
        while self.max_ops_num == numop:
            self.step1(hydrosuper, part)
            hydrosuper.killbas(self.tokill, self.totakeover, self.numops)

            max_bas = part.domainmax(np.max(self.basin_count))
            if part.rank == 0:
                  INFO("Step 1 iter. {0}, maxops: {1}->max_bas : {2}".format(i, self.max_ops_num, max_bas))
            i+=1
            hydrosuper.fetch(part)
            comm.Barrier()
        
        # Step 2 : HTUs going to the same neighbour
        i = 1    
        self.max_ops_num_halo = self.numop; self.max_ops_num_core = self.numop
        while (self.max_ops_num_halo == self.numop) or (self.max_ops_num_core == self.numop):

           # Points with outflow in the core
           if self.max_ops_num_core == self.numop:
               self.step2(hydrosuper, part, "core")
               hydrosuper.killbas(self.tokill, self.totakeover, self.numops)
               max_bas = part.domainmax(np.max(self.basin_count))
               if part.rank == 0:
                  INFO("Step 2 iter. {0} (core) maxops: {1}->max_bas : {2}".format(i, self.max_ops_num_core, max_bas))
               comm.Barrier()
               hydrosuper.fetch(part)
               #comm.Barrier()
           # Points with outflow in the halo

           if self.max_ops_num_halo == self.numop:
               self.step2(hydrosuper, part, "halo")
               hydrosuper.killbas(self.tokill, self.totakeover, self.numops)

               max_bas = part.domainmax(np.max(self.basin_count))
               if part.rank == 0:
                  INFO("Step 2 iter. {0} (halo) maxops: {1}->max_bas : {2}".format(i, self.max_ops_num_halo, max_bas))        
               comm.Barrier()
               hydrosuper.fetch(part)
               #comm.Barrier()
           i += 1 

         
        # Step 3 : HTUs with the same ID          
        i = 1       
        self.max_ops_num_halo = self.numop; self.max_ops_num_core = self.numop
        while (self.max_ops_num_halo == self.numop) or (self.max_ops_num_core == self.numop):
           # points with outflow in the core
           if self.max_ops_num_core == self.numop:
              self.step3(hydrosuper, part, "core")
              hydrosuper.killbas(self.tokill, self.totakeover, self.numops)

              max_bas = part.domainmax(np.max(self.basin_count))
              if part.rank == 0:
                  INFO("Step 3 iter. {0} (core) maxops: {1}->max_bas : {2}".format(i, self.max_ops_num_core, max_bas))       
              hydrosuper.fetch(part)
              comm.Barrier()

           # points with outflow in the halo
           if self.max_ops_num_halo == numop:
               self.step3(hydrosuper, part, "halo")
               hydrosuper.killbas(self.tokill, self.totakeover, self.numops)
               max_bas = part.domainmax(np.max(self.basin_count))
               if part.rank == 0:
                  INFO("Step 3 iter. {0} (halo) maxops: {1}->max_bas : {2}".format(i, self.max_ops_num_halo, max_bas))
               hydrosuper.fetch(part)
               comm.Barrier()
           i += 1         

        
        # Step 4 : HTUs going to the same point
        i = 1
        self.max_ops_num = self.numop
        while (self.max_ops_num == numop):    
            self.step4(hydrosuper, part)
            hydrosuper.killbas(self.tokill, self.totakeover, self.numops)
            max_bas = part.domainmax(np.max(self.basin_count))
            if part.rank == 0:
               INFO("Step 4 iter. {0} (core) maxops: {1}->max_bas : {2}".format(i, self.max_ops_num, max_bas))
            hydrosuper.fetch(part)
            comm.Barrier()
            i += 1

                
        # Step 5 : Brutal method  
        self.step5(hydrosuper, part)
        hydrosuper.killbas(self.tokill, self.totakeover, self.numops)
        max_bas = part.domainmax(np.max(self.basin_count))
        if part.rank == 0:
            INFO("Step 5, maxops: {0}->max_bas : {1}".format(self.numopmax, max_bas))

        hydrosuper.fetch(part)
        comm.Barrier()
        

    ########################
    #
    # METHODs FOR EACH STEP
    #

    def step1(self, hydrosuper, part):
        """ 
        Step 1 : coastal outflow or river outflow
        """
        self.load_data(hydrosuper)
        self.tokill     = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.totakeover = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.numops     = np.zeros((self.nbpt), dtype = np.int32)

        for k in range(len(self.landcorelist)):
            changes = self.get_changes_step1(k)
            if len(changes)>0:
                for ch in changes:
                    orig_ops = self.numops[self.landcorelist[k]]
                    if orig_ops < self.numop:
                        self.add_changes(ch, k)
            self.reindex_changes(k)

        self.max_ops_num = part.domainmax(max(self.numops))
        part.landsendtohalo(self.numops, order='F')
        part.landsendtohalo(self.tokill, order='F')
        part.landsendtohalo(self.totakeover, order='F')

        return   
    #
    def step2(self, hydrosuper, part, neigh_domain):
        """ 
        Step 2 : neighbour
        """
        self.load_data(hydrosuper)
        self.tokill     = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.totakeover = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.numops     = np.zeros((self.nbpt), dtype = np.int32)
        
        for k in range(len(self.landcorelist)):
            changes = self.get_changes_step2(k, neigh_domain)
            if len(changes)>0:
                for ch in changes:
                    orig_ops = self.numops[self.landcorelist[k]]
                    if orig_ops < self.numop:
                        self.add_changes(ch, k)

            self.reindex_changes(k)

        if neigh_domain == "core":
            self.max_ops_num_core = part.domainmax(max(self.numops))
        elif neigh_domain == "halo":
            self.max_ops_num_halo = part.domainmax(max(self.numops))

        part.landsendtohalo(self.numops, order='F')    
        part.landsendtohalo(self.tokill, order='F')
        part.landsendtohalo(self.totakeover, order='F')

        return
    #
    def step3(self, hydrosuper, part, neigh_domain):
        """ 
        Step 3 : Same basin_id
        """ 
        self.load_data(hydrosuper)
        self.tokill     = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.totakeover = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.numops     = np.zeros((self.nbpt), dtype = np.int32)
        
        for k in range(len(self.landcorelist)):
            changes = self.get_changes_step3(k, neigh_domain)
            if len(changes)>0:
                for ch in changes:
                    orig_ops = self.numops[self.landcorelist[k]]
                    if orig_ops < self.numop:
                        self.add_changes(ch, k)
            self.reindex_changes(k)   

        if neigh_domain == "core":
            self.max_ops_num_core = part.domainmax(max(self.numops))
        elif neigh_domain == "halo":
            self.max_ops_num_halo = part.domainmax(max(self.numops))

        part.landsendtohalo(self.numops, order='F')    
        part.landsendtohalo(self.tokill, order='F')
        part.landsendtohalo(self.totakeover, order='F')

        return
    # 
    def step4(self, hydrosuper, part):
        """ 
        Step 4 : htus inside point
        """ 
        self.load_data(hydrosuper)
        self.tokill     = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.totakeover = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.numops     = np.zeros((self.nbpt), dtype = np.int32)
        
        for k in range(len(self.landcorelist)):
            tokk, totakk = self.get_changes_step4(k)

            if len(tokk)>0:
               ops = min(self.numop,len(tokk))
   
               self.tokill[self.landcorelist[k],:ops]  = tokk[:ops]
               self.totakeover [self.landcorelist[k],:ops] = totakk[:ops]
               self.numops[self.landcorelist[k]] = ops 

            self.reindex_changes(k)

        self.max_ops_num = part.domainmax(max(self.numops))
        part.landsendtohalo(self.numops, order='F')    
        part.landsendtohalo(self.tokill, order='F')
        part.landsendtohalo(self.totakeover, order='F')

        return
    #
    def step5(self, hydrosuper, part):
        """ 
        # Step 5 : htus inside point
        """ 
        # Limit of the area of basin to delete by step 5 (in %)
        limit_step5 = 5
        self.load_data(hydrosuper)
        
        # Number of basin to delete in each point of the core
        num2del = []
        for k, k_land in enumerate(self.landcorelist):
            num2del.append(self.basin_count[k_land]-self.nbasmax)
        # Get the maximum over the whole domain
        self.numopmax = part.domainmax(np.max(num2del))

        if self.numopmax>self.numop:
           print("Error - cannot finish to resolve truncate by step 5")
           print("Try with an higher nbasmax")
           sys.exit()


        self.tokill     = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.totakeover = np.zeros((self.nbpt,self.numop), dtype = np.int32)
        self.numops     = np.zeros((self.nbpt), dtype = np.int32)

        for k, k_land in enumerate(self.landcorelist):
            if num2del[k] > 0:
                B = np.argsort(self.fetch[k])[:num2del[k]]
                tokk = [k+1 for k in B]
                totakk = [np.argsort(self.fetch[k])[-1] + 1] * num2del[k]

                area_basin_to_truncate = np.sum([self.basin_area[k][l] for l in B])

                if  (area_basin_to_truncate/np.sum(self.basin_area[k]))*100 < limit_step5:
                    ops = len(tokk)
                    self.tokill[k_land,:ops] = tokk
                    self.totakeover[k_land,:ops] = totakk
                    self.numops[k_land] = ops

                    self.reindex_changes(k)

                else:
                    print("Error - cannot finish to resolve truncate by step 5")
                    print("Try with an higher nbasmax")
                    sys.exit()
        
        part.landsendtohalo(self.numops, order='F')    
        part.landsendtohalo(self.tokill, order='F')
        part.landsendtohalo(self.totakeover, order='F')

        return

    ########################
    #
    # METHOD to GET THE LIST OF CHANGES
    #

    def get_changes_step1(self, npt):
        changes = []

        # outflow_grid = -2
        A = np.where(self.outflow_grid[npt] == -2)[0] 
        if len(A)>1:
            B = self.fetch[npt][A]
            changes.append([A[k]+1 for k in np.argsort(B)])
        # outflow_grid = -3
        A = np.where(self.outflow_grid[npt] == -3)[0] 
        if len(A)>1:
            B = self.fetch[npt][A]
            changes.append([A[k]+1 for k in np.argsort(B)])
        # We are working on core, it's ok to check outflow_grid = 0        
        A = np.where(self.outflow_grid[npt] == 0)[0]
        if len(A)>1:
            B = self.fetch[npt][A]
            changes.append([A[k]+1 for k in np.argsort(B)])

    
        return changes 

    def get_changes_step2(self, npt, neigh_domain):
        """
        neigh_domain = "halo" or "core"
        """
        # List of neighbour which are outflow
        outgrid = self.outflow_grid[npt]
        outgrid = ma.masked_where(outgrid == self.landcorelist[npt], outgrid)
        outgrid = ma.masked_where(outgrid <= 0, outgrid)
        outgrid = ma.unique(outgrid)

        # List of neighbour in core/halo
        core = []; halo = []
        for g in outgrid:
           if g in self.landhalolist_f:
               halo.append(g)
           else:
               core.append(g)

        # Establishing the list of changes
        changes = []        
  
        if neigh_domain == "core":
            for k in core:
                A = np.where(self.outflow_grid[npt] == k)[0]
                if len(A)>1:
                    B = self.fetch[npt][A]
                    changes.append([A[k]+1 for k in np.argsort(B)])

        elif neigh_domain == "halo":
            for k in halo:
                A = np.where(self.outflow_grid[npt] == k)[0]
                if len(A)>1:
                    B = self.fetch[npt][A]
                    changes.append([A[k]+1 for k in np.argsort(B)])

        return changes


    def get_changes_step3(self, npt, neigh_domain):
        """
        neigh_domain = "halo" or "core"
        """
        # List of basin ID included in the grid point 
        basid = self.basin_id[npt]
        basid = ma.unique(basid)
        
        core = []; halo = []

        # For each basin id, list of point with outflow in the core / halo
        for bid in basid:
           htus = np.where(basid == bid)[0]
           outgrid = [self.outflow_grid[npt][htu] for htu in htus]
           c = []; h = []

           for g,htu in zip(outgrid,htus):
               if g in self.landhalolist_f:
                   h.append(htu)
               elif g in self.landcorelist_f:
                   c.append(htu)
           if len(h)>1:
               halo.append(h)
           if len(c)>1:
               core.append(c)

        # Establishing the list of changes 
        changes = []          

        if neigh_domain == "core":
            for c in core:
                B = self.fetch[npt][c]
                changes.append([c[k]+1 for k in np.argsort(B)])

        elif neigh_domain == "halo":
            for h in halo:
                B = self.fetch[npt][h]
                changes.append([h[k]+1 for k in np.argsort(B)])

        return changes

    def get_changes_step4(self, npt):
        tok = []
        totak = []

        # List of HTUs flowing in the same grid
        outgrid = self.outflow_grid[npt]
        htus = np.where(outgrid == self.landcorelist[npt])[0]
        
        # Establishing the lists of changes 
        if len(htus)>0:
            B = self.fetch[npt][htus]
            tok = [htus[k]+1 for k in np.argsort(B)] 
            totak = [self.outflow_basin[npt][htus[k]] for k in np.argsort(B)]

        return tok, totak


    ########################
    #
    # BASIC METHODS
    #
    def load_data(self, hydrosuper): 
        """
        Load the core data from hydrosuper
        """
        self.basin_count = hydrosuper.basin_count
        #
        self.outflow_grid = [hydrosuper.outflow_grid[k,:self.basin_count[k]] for k in self.landcorelist]
        self.outflow_basin = [hydrosuper.outflow_basin[k,:self.basin_count[k]] for k in self.landcorelist]
        self.basin_id = [hydrosuper.basin_id[k,:self.basin_count[k]] for k in self.landcorelist]
        self.basin_area = [hydrosuper.basin_area[k,:self.basin_count[k]] for k in self.landcorelist]
        self.fetch = [hydrosuper.fetch_basin[k,:self.basin_count[k]] for k in self.landcorelist]
       
    def add_changes(self, HTUs_to_truncate, k):
        """
        Add the changes to (tokill, totakeover, numops) from HTUs_to_truncate
        HTUs_to_truncate : List of HTUs to truncate, ordered by fetch
        """
        orig_ops = self.numops[self.landcorelist[k]]
        poss_ops = len(HTUs_to_truncate)-1
        ops = min(poss_ops,self.numop-orig_ops)

        self.tokill[self.landcorelist[k],orig_ops:orig_ops+ops] = HTUs_to_truncate[:ops]
        self.totakeover[self.landcorelist[k],orig_ops:orig_ops+ops] = HTUs_to_truncate[1:ops+1]
        self.numops[self.landcorelist[k]] += ops

    def reindex_changes(self, k):
        """
        Reindex the (tokill, totakeover) to anticipate the changes made before
        """
        ops = self.numops[self.landcorelist[k]]
        for l in range(ops-1):
            tok = self.tokill[self.landcorelist[k],l]
            tokill_up     = self.tokill[self.landcorelist[k],l+1:]
            totakeover_up = self.totakeover[self.landcorelist[k],l+1:]
                    
            tokill_up[tokill_up > tok]         -= 1
            totakeover_up[totakeover_up > tok] -= 1

            self.tokill[self.landcorelist[k],l+1:]     = tokill_up
            self.totakeover[self.landcorelist[k],l+1:] = totakeover_up




