#
#
#
import numpy as np
import os, sys
from mpi4py import MPI
import gc
#
import matplotlib as mpl
mpl.use('Agg')
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import time
#
# Gert the information from the configuration file.
#
import configparser
config=configparser.ConfigParser()
config.read("run.def")
nbasmax=config.getint("OverAll", "nbasmax")
numop=config.getint("OverAll", "numop", fallback=100)
OutGraphFile=config.get("OverAll","GraphFile")
DumpHydroSuper=config.getboolean("OverAll","DumpHydroSuper",fallback=False)
wfile=config.get("OverAll","WeightFile",fallback="Weights.nc")
#
import ModelGrid as MG
import Partition as PA
import RPPtools as RPP
import HydroGrid as HG
#
# Logging in MPI : https://groups.google.com/forum/#!topic/mpi4py/SaNzc8bdj6U
#
import getargs
log_master, log_world = getargs.getLogger()
INFO, DEBUG, ERROR = log_master.info, log_master.debug, log_world.error
INFO_ALL, DEBUG_ALL = log_world.info, log_world.debug
#
# Read full grid and partition the domain.
#
gg = MG.GlobalGrid()
#
comm = MPI.COMM_WORLD
szhalo=0
nbcore=comm.Get_size()
rank=comm.Get_rank()
#
# Region of grid to be treated
#
part = PA.partition(gg.ni, gg.nj, gg.land, comm, nbcore, szhalo, rank)

INFO("rank:{0}-nbland:{1}".format(part.rank, part.nbland))
#

with open("./Weight_proc_{0}.txt".format(part.rank), "a") as foo:
   foo.write("Start Weights on rank: {0}\n".format(part.rank))
#
modelgrid=MG.ModelGrid(part.ihstart+gg.igstart,part.nih,part.jhstart+gg.jgstart,part.njh)
#
hydrogrid=HG.HydroGrid(modelgrid.box_land, wfile)
#
# Computes weights of overlap of modelgrid and hydrogrid
#

with open("./Weight_proc_{0}.txt".format(part.rank), "a") as foo:
   foo.write("Proc {0}: size core : {1}\n".format(part.rank, len(part.landcorelist)))

w = RPP.compweights(wfile, part, modelgrid, hydrogrid)
#
#

